#pragma once

#include "type_info.h"

#include <iostream>
#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <typeinfo>
#include <string.h>
#include "smart_assert.h"

template<typename T>
class Data3D {
public:
    Data3D( const cv::Vec3i& n_size = cv::Vec3i(0,0,0));
    Data3D( const int width, const int height, const int depth );
    Data3D( const cv::Vec3i& n_size, const T& value );
    template <class T2>
    Data3D( const Data3D<T2>& src );
    const Data3D<T>& operator=( const Data3D<T>& src );

    virtual ~Data3D(void) { }

    inline void reset( void )
    {
        memset( _mat.data, 0, _size_total * sizeof(T) );
    }
    inline void reset( const cv::Vec3i& n_size )
    {
        _size = n_size;
        long size_slice = _size[0] * _size[1];
        _size_total = size_slice * _size[2];
        _mat = cv::Mat_<T>( _size[2], size_slice );
        memset( _mat.data, 0, _size_total * sizeof(T) );
    }
    inline void reset( const int width, const int height, const int depth )
    {
        _size = cv::Vec3i(width, height, depth);
        long size_slice = _size[0] * _size[1];
        _size_total = size_slice * _size[2];
        _mat = cv::Mat_<T>( _size[2], size_slice );
        memset( _mat.data, 0, _size_total * sizeof(T) );
    }
    void reset( const cv::Vec3i& n_size, const T& value )
    {
        resize( n_size );
        for( cv::MatIterator_<T> it=_mat.begin(); it<_mat.end(); it++ )
        {
            (*it) = value;
        }
    }

    void resize( const cv::Vec3i& n_size )
    {
        _size = n_size;
        long size_slice = _size[0] * _size[1];
        _size_total = size_slice * _size[2];
        _mat = cv::Mat_<T>( _size[2], size_slice );
    }

    inline const int& SX(void) const
    {
        return _size[0];
    }
    inline const int& SY(void) const
    {
        return _size[1];
    }
    inline const int& SZ(void) const
    {
        return _size[2];
    }
    inline const int& get_width(void)  const
    {
        return _size[0];
    }
    inline const int& get_height(void) const
    {
        return _size[1];
    }
    inline const int& get_depth(void)  const
    {
        return _size[2];
    }
    inline const int& get_size_x(void) const
    {
        return _size[0];
    }
    inline const int& get_size_y(void) const
    {
        return _size[1];
    }
    inline const int& get_size_z(void) const
    {
        return _size[2];
    }
    inline const int& get_size(int i) const
    {
        return _size[i];
    }
    inline const long& get_size_total(void) const
    {
        return _size_total;
    }
    inline const cv::Vec3i& get_size(void) const
    {
        return _size;
    }
    inline const bool is_empty( void ) const
    {
        return get_size_total()==0;
    }

    virtual const inline T& at( const int& i ) const
    {
        return _mat( i );
    }
    virtual inline T& at( const int& i )
    {
        return _mat( i );
    }
    virtual inline const T& at( const int& x, const int& y, const int& z ) const
    {
        return _mat( z, y * _size[0] + x );
    }
    virtual inline T& at( const int& x, const int& y, const int& z )
    {
        return _mat( z, y * _size[0] + x );
    }
    virtual inline const T& at( const cv::Vec3i& pos ) const
    {
        return at( pos[0], pos[1], pos[2] );
    }
    virtual inline T& at( const cv::Vec3i& pos )
    {
        return at( pos[0], pos[1], pos[2] );
    }

    inline cv::Mat_<T>& getMat()
    {
        return _mat;
    }
    inline const cv::Mat_<T>& getMat() const
    {
        return _mat;
    }
    inline cv::Mat_<T> getMat( int slice ) const
    {
        return _mat.row(slice).reshape( 0, get_height() ).clone();
    }

    void SetFromArray( const T* array );
    void CopyToArray( T* array );

    template<typename T1, typename T2>
    friend const Data3D<T1>& operator*=( Data3D<T1>& left, const T2& right );
    template<typename T1, typename T2, typename T3>
    friend bool subtract3D( const Data3D<T1>& src1, const Data3D<T2>& src2, Data3D<T3>& dst );
    template<typename T1, typename T2, typename T3>
    friend bool multiply3D( const Data3D<T1>& src1, const Data3D<T2>& src2, Data3D<T3>& dst );

    cv::Vec<T, 2> get_min_max_value() const
    {
        cv::Vec<T, 2> min_max;
        cv::Point minLoc, maxLoc;
        cv::minMaxLoc( _mat, NULL, NULL, &minLoc, &maxLoc);
        min_max[0] = _mat( minLoc );
        min_max[1] = _mat( maxLoc );
        return min_max;
    }

public:
    /* copy data of dimension i to a new Image3D structure
       e.g. this is useful when trying to visualize vesselness which is
       a multidimensional data structure
       Yuchen: p.s. I have no idea how to move the funciton body out the
       class definition nicely. Let me figure it later maybe. */
    template<typename T2>
    void copyDimTo( Data3D<T2>& dst, int dim ) const
    {
        dst.reset( this->get_size() );
        int x, y, z;
        for( z=0; z<get_size_z(); z++ ) for( y=0; y<get_size_y(); y++ ) for( x=0; x<get_size_x(); x++ ) {
                    dst.at(x,y,z) = this->at(x,y,z)[dim];
                }
    }


    // return true if a index is valide for the data
    inline bool isValid( const int& x, const int& y, const int& z ) const
    {
        return ( x>=0 && x<get_size_x() &&
                 y>=0 && y<get_size_y() &&
                 z>=0 && z<get_size_z() );
    }
    inline bool isValid( const cv::Vec3i& v ) const
    {
        return isValid( v[0], v[1], v[2] );
    }

    inline const T* getData(void) const
    {
        return (T*) getMat().data;
    }

    inline T* getData(void)
    {
        return (T*) getMat().data;
    }

protected:
    cv::Vec3i _size;
    long _size_total;
    cv::Mat_<T> _mat;
};


// Default Constructor
template <typename T>
Data3D<T>::Data3D( const cv::Vec3i& n_size )
{
    reset(n_size);
}
template <typename T>
Data3D<T>::Data3D( const int width, const int height, const int depth )
{
    reset(width, height, depth);
}

// Constructor with size and value
template <typename T>
Data3D<T>::Data3D( const cv::Vec3i& n_size, const T& value )
{
    reset(n_size, value);
}


// Copy Constructor
template<typename T>
template<typename T2>
Data3D<T>::Data3D( const Data3D<T2>& src )
{
    // resize
    this->resize( src.get_size() );
    // copy the data over
    cv::MatIterator_<T>       this_it;
    cv::MatConstIterator_<T2> src_it;
    for( this_it = this->getMat().begin(), src_it = src.getMat().begin();
            this_it < this->getMat().end(),   src_it < src.getMat().end();
            this_it++, src_it++ ) {
        // Yuchen: The convertion from T2 to T may be unsafe
        *(this_it) = T( *(src_it) );
    }
}


template<typename T>
const Data3D<T>& Data3D<T>::operator=( const Data3D<T>& src )
{
    // resize
    this->resize( src.get_size() );

    // copy the data over
    cv::MatIterator_<T>       this_it;
    cv::MatConstIterator_<T> src_it;
    for( this_it = this->getMat().begin(), src_it = src.getMat().begin();
            this_it < this->getMat().end(),   src_it < src.getMat().end();
            this_it++, src_it++ ) {
        *(this_it) = *(src_it);
    }

    return *this;
}


template <typename T>
void Data3D<T>::SetFromArray( const T* array )
{
    memcpy(_mat.data, array, _size_total * sizeof(T));
}

template <typename T>
void Data3D<T>::CopyToArray( T* array )
{
    memcpy(array, _mat.data, _size_total * sizeof(T));
}

//////////////////////////////////////////////////////////////////////
// Operator Overloading and other friend functions

template<typename T, typename T2>
const Data3D<T>& operator*=( Data3D<T>& left, const T2& right )
{
    left._mat*=right;
    return left;
}

template<typename T1, typename T2, typename T3>
bool subtract3D( const Data3D<T1>& src1, const Data3D<T2>& src2, Data3D<T3>& dst )
{
    smart_return( src1.get_size()==src2.get_size(), "Sizes should match.", false);

    if( dst.get_size() != src1.get_size() ) {
        dst.reset( src1.get_size() );
    }

    int x, y, z;
    for( z=0; z<src1.get_size_z(); z++ ) {
        for( y=0; y<src1.get_size_y(); y++ ) {
            for( x=0; x<src1.get_size_x(); x++ ) {
                dst.at(x,y,z) = T3(src1.at(x,y,z) - src2.at(x,y,z));
            }
        }
    }
    return true;
}


template<typename T1, typename T2, typename T3>
bool multiply3D( const Data3D<T1>& src1, const Data3D<T2>& src2, Data3D<T3>& dst )
{
    smart_return_false( src1.get_size()==src2.get_size(),
                        "Source sizes are supposed to be cv::Matched.");

    if( dst.get_size() != src1.get_size() ) {
        dst.reset( src1.get_size() );
    }

    int x, y, z;
    for( z=0; z<src1.get_size_z(); z++ ) {
        for( y=0; y<src1.get_size_y(); y++ ) {
            for( x=0; x<src1.get_size_x(); x++ ) {
                dst.at(x,y,z) = src1.at(x,y,z) * src2.at(x,y,z);
            }
        }
    }
    return true;
}
