#include "eigen.h"

/* Compute the eigen value decomposition of a 3 by 3 symetric matrix
Input:
A = [ A11  A12  A13     = [ A[0]  A[1]  A[2]
A21  A22  A23          0    A[3]  A[4]
A31  A32  A33 ];       0     0    A[5] ];
Output:
eigenvalues, eigenvectors

Reference: http://en.wikipedia.org/wiki/Eigenvalue_algorithm#3.C3.973_matrices
*/
#include <cmath>
#include <assert.h>
#include <algorithm>
#include <iostream>
// #include <opencv2/core/types_c.h>
// #include <opencv2/core/core_c.h>


#if !defined(SmallEpsilon) && !defined(BigEpsilon)
#define SmallEpsilon 1.0e-18
#define BigEpsilon 1.0e-4
#endif















#include "math.h"
#ifdef MAX
#undef MAX
#endif
#define MAX(a, b) ((a)>(b)?(a):(b))
#define n 3

// address/from/source: https://github.com/fethallah/neurons/blob/master/frangi_filter_version2a/eig3volume.c

/* Eigen decomposition code for symmetric 3x3 matrices, copied from the public
 * domain Java Matrix library JAMA. */
static float hypot2(float x, float y) { return (float) sqrt(x*x+y*y); }

__inline float absd(float val){ if(val>0){ return val;} else { return -val;} };

/* Symmetric Householder reduction to tridiagonal form. */
static void tred2(float V[n][n], float d[n], float e[n]) {

/*  This is derived from the Algol procedures tred2 by */
/*  Bowdler, Martin, Reinsch, and Wilkinson, Handbook for */
/*  Auto. Comp., Vol.ii-Linear Algebra, and the corresponding */
/*  Fortran subroutine in EISPACK. */

//     std::cout << "Entering tred2" << std::endl;

    int i, j, k;
    float scale;
    float f, g, h;
    float hh;
    for (j = 0; j < n; j++) {d[j] = V[n-1][j]; }

    /* Householder reduction to tridiagonal form. */

    for (i = n-1; i > 0; i--) {
        /* Scale to avoid under/overflow. */
        scale = 0.0;
        h = 0.0;
        for (k = 0; k < i; k++) { scale = scale + fabs(d[k]); }
        if (scale == 0.0) {
            e[i] = d[i-1];
            for (j = 0; j < i; j++) { d[j] = V[i-1][j]; V[i][j] = 0.0;  V[j][i] = 0.0; }
        } else {

            /* Generate Householder vector. */

            for (k = 0; k < i; k++) { d[k] /= scale; h += d[k] * d[k]; }
            f = d[i-1];
            g = sqrt(h);
            if (f > 0) { g = -g; }
            e[i] = scale * g;
            h = h - f * g;
            d[i-1] = f - g;
            for (j = 0; j < i; j++) { e[j] = 0.0; }

            /* Apply similarity transformation to remaining columns. */

            for (j = 0; j < i; j++) {
                f = d[j];
                V[j][i] = f;
                g = e[j] + V[j][j] * f;
                for (k = j+1; k <= i-1; k++) { g += V[k][j] * d[k]; e[k] += V[k][j] * f; }
                e[j] = g;
            }
            f = 0.0;
            for (j = 0; j < i; j++) { e[j] /= h; f += e[j] * d[j]; }
            hh = f / (h + h);
            for (j = 0; j < i; j++) { e[j] -= hh * d[j]; }
            for (j = 0; j < i; j++) {
                f = d[j]; g = e[j];
                for (k = j; k <= i-1; k++) { V[k][j] -= (f * e[k] + g * d[k]); }
                d[j] = V[i-1][j];
                V[i][j] = 0.0;
            }
        }
        d[i] = h;
    }

    /* Accumulate transformations. */

    for (i = 0; i < n-1; i++) {
        V[n-1][i] = V[i][i];
        V[i][i] = 1.0;
        h = d[i+1];
        if (h != 0.0) {
            for (k = 0; k <= i; k++) { d[k] = V[k][i+1] / h;}
            for (j = 0; j <= i; j++) {
                g = 0.0;
                for (k = 0; k <= i; k++) { g += V[k][i+1] * V[k][j]; }
                for (k = 0; k <= i; k++) { V[k][j] -= g * d[k]; }
            }
        }
        for (k = 0; k <= i; k++) { V[k][i+1] = 0.0;}
    }
    for (j = 0; j < n; j++) { d[j] = V[n-1][j]; V[n-1][j] = 0.0; }
    V[n-1][n-1] = 1.0;
    e[0] = 0.0;

//     std::cout << "Leaving tred2" << std::endl;
}

/* Symmetric tridiagonal QL algorithm. */
static void tql2(float V[n][n], float d[n], float e[n]) {

/*  This is derived from the Algol procedures tql2, by */
/*  Bowdler, Martin, Reinsch, and Wilkinson, Handbook for */
/*  Auto. Comp., Vol.ii-Linear Algebra, and the corresponding */
/*  Fortran subroutine in EISPACK. */

//     std::cout << "Entering tql2" << std::endl;

    int i, j, k, l, m;
    float f;
    float tst1;
    float eps;
    int iter;
    float g, p, r;
    float dl1, h, c, c2, c3, el1, s, s2;

    for (i = 1; i < n; i++) { e[i-1] = e[i]; }
    e[n-1] = 0.0;

    f = 0.0;
    tst1 = 0.0;
    eps = pow(2.0, -52.0);
    for (l = 0; l < n; l++) {

        /* Find small subdiagonal element */

        tst1 = MAX(tst1, fabs(d[l]) + fabs(e[l]));
        m = l;
        while (m < n) {
            if (fabs(e[m]) <= eps*tst1) { break; }
            m++;
        }

        /* If m == l, d[l] is an eigenvalue, */
        /* otherwise, iterate. */

        if (m > l) {
            iter = 0;
            do {
                iter = iter + 1;  /* (Could check iteration count here.) */
                /* Compute implicit shift */
                g = d[l];
                p = (d[l+1] - g) / (2.0 * e[l]);
                r = hypot2(p, 1.0);
                if (p < 0) { r = -r; }
                d[l] = e[l] / (p + r);
                d[l+1] = e[l] * (p + r);
                dl1 = d[l+1];
                h = g - d[l];
                for (i = l+2; i < n; i++) { d[i] -= h; }
                f = f + h;
                /* Implicit QL transformation. */
                p = d[m]; c = 1.0; c2 = c; c3 = c;
                el1 = e[l+1]; s = 0.0; s2 = 0.0;
                for (i = m-1; i >= l; i--) {
                    c3 = c2;
                    c2 = c;
                    s2 = s;
                    g = c * e[i];
                    h = c * p;
                    r = hypot2(p, e[i]);
                    e[i+1] = s * r;
                    s = e[i] / r;
                    c = p / r;
                    p = c * d[i] - s * g;
                    d[i+1] = h + s * (c * g + s * d[i]);
                    /* Accumulate transformation. */
                    for (k = 0; k < n; k++) {
                        h = V[k][i+1];
                        V[k][i+1] = s * V[k][i] + c * h;
                        V[k][i] = c * V[k][i] - s * h;
                    }
                }
                p = -s * s2 * c3 * el1 * e[l] / dl1;
                e[l] = s * p;
                d[l] = c * p;

                /* Check for convergence. */
            } while (fabs(e[l]) > eps*tst1);
        }
        d[l] = d[l] + f;
        e[l] = 0.0;
    }

    /* Sort eigenvalues and corresponding vectors. */
    for (i = 0; i < n-1; i++) {
        k = i;
        p = d[i];
        for (j = i+1; j < n; j++) {
            if (d[j] < p) {
                k = j;
                p = d[j];
            }
        }
        if (k != i) {
            d[k] = d[i];
            d[i] = p;
            for (j = 0; j < n; j++) {
                p = V[j][i];
                V[j][i] = V[j][k];
                V[j][k] = p;
            }
        }
    }

//     std::cout << "Leaving tql2" << std::endl;
}

void eigen_decomposition(float A[n][n], float V[n][n], float d[n]) {
    float e[n];
    float da[3];
    float dt, dat;
    float vet[3];
    int i, j;

//     std::cout << "Entering egien_decompositiom" << std::endl;

    for (i = 0; i < n; i++) {
        for (j = 0; j < n; j++) {
            V[i][j] = A[i][j];
        }
    }
    tred2(V, d, e);
    tql2(V, d, e);

    /* Sort the eigen values and vectors by abs eigen value */
    da[0]=absd(d[0]); da[1]=absd(d[1]); da[2]=absd(d[2]);
    if((da[0]>=da[1])&&(da[0]>da[2]))
    {
        dt=d[2];   dat=da[2];    vet[0]=V[0][2];    vet[1]=V[1][2];    vet[2]=V[2][2];
        d[2]=d[0]; da[2]=da[0];  V[0][2] = V[0][0]; V[1][2] = V[1][0]; V[2][2] = V[2][0];
        d[0]=dt;   da[0]=dat;    V[0][0] = vet[0];  V[1][0] = vet[1];  V[2][0] = vet[2];
    }
    else if((da[1]>=da[0])&&(da[1]>da[2]))
    {
        dt=d[2];   dat=da[2];    vet[0]=V[0][2];    vet[1]=V[1][2];    vet[2]=V[2][2];
        d[2]=d[1]; da[2]=da[1];  V[0][2] = V[0][1]; V[1][2] = V[1][1]; V[2][2] = V[2][1];
        d[1]=dt;   da[1]=dat;    V[0][1] = vet[0];  V[1][1] = vet[1];  V[2][1] = vet[2];
    }
    if(da[0]>da[1])
    {
        dt=d[1];   dat=da[1];    vet[0]=V[0][1];    vet[1]=V[1][1];    vet[2]=V[2][1];
        d[1]=d[0]; da[1]=da[0];  V[0][1] = V[0][0]; V[1][1] = V[1][0]; V[2][1] = V[2][0];
        d[0]=dt;   da[0]=dat;    V[0][0] = vet[0];  V[1][0] = vet[1];  V[2][0] = vet[2];
    }

//     std::cout << "Leaving egien_decompositiom" << std::endl;
}
















/*
template<typename T=float>
inline void normalize( T v[3] )
{
    T length = 0;
    for( int i=0; i<3; i++ ) {
        length += v[i] * v[i];
    }
//     assert( length > SmallEpsilon && "Cannot normalized zero vector" );
    if ( length <= SmallEpsilon )
        return;

    length = sqrt( length );
    for( int i=0; i<3; i++ ) {
        v[i] /= length;
    }
}

template<typename T=float>
inline void cross_product( const T v1[3], const T v2[3], T v3[3] )
{
    v3[0] = v1[1] * v2[2] - v1[2] * v2[1];
    v3[1] = v1[2] * v2[0] - v1[0] * v2[2];
    v3[2] = v1[0] * v2[1] - v1[1] * v2[0];
    normalize( v3 );
}

template<typename T=float>
void normal_vectors(  const T v1[3], T v2[3], T v3[3] )
{
    // an arbitrary combination of the following two vectors
    // is a normal to v1
    // alpha * (-v1[1], v1[0], 0) + beta * (0, -v1[2], v[1])

    if( abs(v1[0])>BigEpsilon || abs(v1[1])>BigEpsilon ) {
        v2[0] = -v1[1];
        v2[1] =  v1[0];
        v2[2] =  0;
    }
    else {
        v2[0] =  0;
        v2[1] = -v1[2];
        v2[2] =  v1[1];
    }

    normalize( v2 );

    cross_product( v1, v2, v3 );
}
*/

template<typename T=float>
void eigen_decomp( const T A[6], T eigenvalues[3], T eigenvectors[3][3] )
{
    float Asym[n][n];
    Asym[0][0] = A[0]; Asym[0][1] = A[1]; Asym[0][2] = A[2];
    Asym[1][0] = A[1]; Asym[1][1] = A[3]; Asym[1][2] = A[4];
    Asym[2][0] = A[2]; Asym[2][1] = A[4]; Asym[2][2] = A[5];

    /* Function below sorts the eigen values and vectors by abs eigen value */
    eigen_decomposition(Asym, eigenvectors, eigenvalues);

    return;

//     static CvMat mat, *eigenVec, *eigenVal;
//     static float Ain1D[9];
//
//     int cnt = 0;
//     Ain1D[cnt++] = A[0];
//     Ain1D[cnt++] = A[1];
//     Ain1D[cnt++] = A[2];
//     Ain1D[cnt++] = A[1];
//     Ain1D[cnt++] = A[3];
//     Ain1D[cnt++] = A[4];
//     Ain1D[cnt++] = A[2];
//     Ain1D[cnt++] = A[4];
//     Ain1D[cnt++] = A[5];
//
//     int i;
//     std::cout << std::endl;
//     for (i = 0; i < 9; i++) {
//         std::cout << Ain1D[i] << " ";
//         if ((i == 1) || (i == 4))
//             std::cout << std::endl;
//     }
//     std::cout << std::endl << std::endl;
//
//     mat = cvMat(3, 3, CV_32FC1, Ain1D);
//
//     cvEigenVV(&mat, eigenVec, eigenVal, 1e-300);
//
//     for(int i=0;i < 3;i++){
//         eigenvalues[i] = cvmGet(eigenVal,i,0); //Fetching Eigen Value
//
//         for(int j=0;j < 3;j++){
//             eigenvectors[i][j] = cvmGet(eigenVec,i,j); //Fetching each component of Eigenvector i
//         }
//     }
//
//     return;

/*
    const T& A11 = A[0];
    const T& A12 = A[1];
    const T& A13 = A[2];
    const T& A22 = A[3];
    const T& A23 = A[4];
    const T& A33 = A[5];

    T& eig1 = eigenvalues[0];
    T& eig2 = eigenvalues[1];
    T& eig3 = eigenvalues[2];

    T p1 = A12 * A12 + A13 * A13 + A23 * A23;
    if( p1 < SmallEpsilon ) { // if A is diagonal
        eig1 = A11;
        eig2 = A22;
        eig3 = A33;
        memset( eigenvectors, 0, sizeof(T) * 9 );
        eigenvectors[0][0] = eigenvectors[1][1] = eigenvectors[2][2] = 1;
    }
    else {   // if A is not diagonal
        // Compute 1/3 of the trace of matrix A: trace(A)/3
        T q = ( A11 + A22 + A33 ) / 3;
        T p2 = (A11-q)*(A11-q) + (A22-q)*(A22-q) + (A33-q)*(A33-q) + 2 * p1;
        T p = sqrt( p2 / 6);

        // Construct matrix B
        // B = (1 / p) * (A - q * I), where I is the identity matrix
        T B11 = (1 / p) * (A11-q);
        T B12 = (1 / p) * A12;
        T& B21 = B12;
        T B13 = (1 / p) * A13;
        T& B31 = B13;
        T B22 = (1 / p) * (A22-q);
        T B23 = (1 / p) * A23;
        T& B32 = B23;
        T B33 = (1 / p) * (A33-q);

        // Determinant of a 3 by 3 matrix B
        // Reference: http://www.mathworks.com/help/aeroblks/determinantof3x3matrix.html
        T detB = B11*(B22*B33-B23*B32) - B12*(B21*B33-B23*B31) + B13*(B21*B32-B22*B31);

        // In exact arithmetic for a symmetric matrix  -1 <= r <= 1
        // but computation error can leave it slightly outside this range.
        T r = detB / 2;
        T phi;
        const T M_PI3 = T(3.14159265 / 3);
        if( r <= -1.0f ) {
            phi = M_PI3;
        }
        else if (r >= 1.0f)
            phi = 0;
        else {
            phi = acos(r) / 3;
        }

        // The eigenvalues satisfy eig3 <= eig2 <= eig1
        // Notice that: trace(A) = eig1 + eig2 + eig3
        eig1 = q + 2 * p * cos( phi );
        eig3 = q + 2 * p * cos( phi + 2 * M_PI3 );
        eig2 = 3 * q - eig1 - eig3;

        // make sure that |eig1| >= |eig2| >= |eig3|
        if( abs(eig1) < abs(eig2) ) std::swap(eig1, eig2);
        if( abs(eig2) < abs(eig3) ) std::swap(eig2, eig3);
        if( abs(eig1) < abs(eig2) ) std::swap(eig1, eig2);

        // If eig1 is too small, it is not worth to compute the eigenvectors
        if( abs(eig1) < BigEpsilon ) {
            memset( eigenvectors, 0, sizeof(T) * 9 );
            eigenvectors[0][0] = eigenvectors[1][1] = eigenvectors[2][2] = 1;
        }

        // Compute the corresponding eigenvectors
        // Reference: [Cayley-Hamilton_theorem](http://en.wikipedia.org/wiki/Cayley-Hamilton_theorem)
        // If eig1, eig2, eig3 are the distinct eigenvalues of the matrix A;
        // that is: eig1 != eig2 != eig3. Then
        // (A - eig1 * I)(A - eig2 * I)(A - eig3 * I) = 0
        // Thus the columns of the product of any two of these matrices
        // will contain an eigenvector for the third eigenvalue.
        else if( abs(eig1-eig2)<BigEpsilon && abs(eig2-eig3)<BigEpsilon ) {
            memset( eigenvectors, 0, sizeof(T) * 9 );
            eigenvectors[0][0] = eigenvectors[1][1] = eigenvectors[2][2] = 1;
        }
        else if ( abs(eig1-eig2)<BigEpsilon ) {
            // tested
            eigenvector_from_value( A, eig1, eig2, eigenvectors[2] );
            normal_vectors( eigenvectors[2], eigenvectors[1], eigenvectors[0] );
        }
        else if ( abs(eig2-eig3)<BigEpsilon ) {
            // tested
            eigenvector_from_value( A, eig2, eig3, eigenvectors[0] );
            normal_vectors( eigenvectors[0], eigenvectors[1], eigenvectors[2] );
        }
        else {
            // eig1!=eig2 && eig2!=eig3 && eig1!=eig3
            // tested
            eigenvector_from_value( A, eig2, eig3, eigenvectors[0] );
            eigenvector_from_value( A, eig1, eig3, eigenvectors[1] );
            eigenvector_from_value( A, eig1, eig2, eigenvectors[2] );
        }
    }
*/
}

/*
template<typename T=float>
inline void eigenvector_from_value( const T A[6],
                                    const T& eig1, const T& eig2, T eigenvector[3] )
{
    const T& A11 = A[0];
    const T& A12 = A[1];
    const T& A13 = A[2];
    const T& A22 = A[3];
    const T& A23 = A[4];
    const T& A33 = A[5];

    T tempLength = 0, length = 0;
    T tempVector[3];
    T colum[3];
    for( int i=0; i<3; i++ ) {
        switch( i ) {
        case 0:
            colum[0] = A11 - eig2;
            colum[1] = A12;
            colum[2] = A13;
            break;
        case 1:
            colum[0] = A12;
            colum[1] = A22 - eig2;
            colum[2] = A23;
            break;
        default:
            colum[0] = A13;
            colum[1] = A23;
            colum[2] = A33 - eig2;
            break;
        }
        tempVector[0] = colum[0] * (A11 - eig1) + colum[1] * A12 + colum[2] * A13;
        tempVector[1] = colum[0] * A12 + colum[1] * (A22 - eig1) + colum[2] * A23;
        tempVector[2] = colum[0] * A13 + colum[1] * A23 + colum[2] * (A33 - eig1);

        tempLength = 0;
        for( int k=0; k<3; k++ ) {
            tempLength += tempVector[k] * tempVector[k];
        }

        // There are three columns in each resulting matrix of ( A-lambda_i * I ) * ( A-lambda_j * I )
        // We choose the one with the maximum length for the sake of accuracy.
        if( length < tempLength ) {
            length = tempLength;
            memcpy( eigenvector, tempVector, sizeof(T)*3);
        }

        if( length > BigEpsilon ) break;
    }

    // The vector is almost zero,
//     assert( length >= SmallEpsilon && "All eigenvector are zero. " );

    // TODO: for debuging
//     if( length < SmallEpsilon ) {
//         std::cerr << "Lenght of vector is too small. Maybe problematic. " << std::endl;
//     }

    // nromailized the vector
    length = sqrt( length );
    for( int k=0; k<3; k++ ) eigenvector[k] /= length;
}
*/
