#include <opencv2/core/core.hpp>
#include <queue>
#include <iostream>
#include "types.h"

#include "processing.h"
#include "processing_code.h"

using namespace cv;
using namespace std;

void ImageProcessing::non_max_suppress( const Data3D<Vesselness_Sig>& src, Data3D<Vesselness_Sig>& dst )
{
    dst.reset( src.get_size() );

    static const float s2 = sqrt(1/2.0f);
    static const float s3 = sqrt(1/3.0f);

    // 13 major orientations in 3d
    static const int MAJOR_DIR_NUM = 13;
    static const Vec3f major_dirs[MAJOR_DIR_NUM] = {
        // directions along the axis, there are 3 of them
        Vec3f(1, 0, 0), Vec3f(0, 1, 0), Vec3f(0, 0, 1),
        // directions that lie within the 3D plane of two axis, there are 6 of them
        Vec3f(s2, s2,0), Vec3f(0,s2, s2), Vec3f(s2,0, s2),
        Vec3f(s2,-s2,0), Vec3f(0,s2,-s2), Vec3f(s2,0,-s2),
        // directions that are equally in between three axis, there are 4 of them
        Vec3f(s3,s3,s3), Vec3f(s3,s3,-s3), Vec3f(s3,-s3,s3), Vec3f(s3,-s3,-s3)
    };

    // there are 26 directions in 3D
    static const int NUM_DIR_3D = 26;
    Vec3i neighbor3d[NUM_DIR_3D] = {
        // directions along the axis, there are 6 of them
        Vec3i(0,0, 1), Vec3i(0, 1,0),  Vec3i( 1,0,0),
        Vec3i(0,0,-1), Vec3i(0,-1,0),  Vec3i(-1,0,0),
        // directions that lie within the 3D plane of two axis, there are 12 of them
        Vec3i(0,1,1), Vec3i(0, 1,-1), Vec3i( 0,-1,1), Vec3i( 0,-1,-1),
        Vec3i(1,0,1), Vec3i(1, 0,-1), Vec3i(-1, 0,1), Vec3i(-1, 0,-1),
        Vec3i(1,1,0), Vec3i(1,-1, 0), Vec3i(-1, 1,0), Vec3i(-1,-1, 0),
        // directions that are equally in between three axis, there are 8 of them
        Vec3i( 1,1,1), Vec3i( 1,1,-1), Vec3i( 1,-1,1), Vec3i( 1,-1,-1),
        Vec3i(-1,1,1), Vec3i(-1,1,-1), Vec3i(-1,-1,1), Vec3i(-1,-1,-1),
    };

    // cross section for the major orientation
    vector<Vec3i> cross_section[MAJOR_DIR_NUM];

    // Setting offsets that are perpendicular to dirs
    for( int i=0; i<MAJOR_DIR_NUM; i++ ) {
        for( int j=0; j<NUM_DIR_3D; j++ ) {
            // multiply the two directions
            float temp = major_dirs[i].dot( neighbor3d[j] );
            // the temp is 0, store this direciton
            if( abs(temp)<1.0e-5 ) {
                cross_section[i].push_back( neighbor3d[j] );
            }
        }
    }

    int x,y,z;
    for( z=0; z<src.get_size_z(); z++ ) {
        for( y=0; y<src.get_size_y(); y++ ) {
            for( x=0; x<src.get_size_x(); x++ ) {
                // non-maximum surpression
                bool isMaximum = true;

                // Method I:
                // find the major orientation
                // assigning the orientation to one of the 13 categories
                const Vec3f& cur_dir = src.at(x,y,z).dir;
                int mdi = 0; // major direction id
                float max_dot_product = 0;
                for( int di=0; di<MAJOR_DIR_NUM; di++ ) {
                    float current_dot_product = abs( cur_dir.dot(major_dirs[di]) );
                    if( max_dot_product < current_dot_product ) {
                        max_dot_product = current_dot_product;
                        mdi = di;// update the major direction id
                    }
                }
                for( unsigned int i=0; i<cross_section[mdi].size(); i++ ) {
                    int ox = x + cross_section[mdi][i][0];
                    int oy = y + cross_section[mdi][i][1];
                    int oz = z + cross_section[mdi][i][2];
                    if( src.isValid(ox,oy,oz) && src.at(x,y,z).rsp < src.at(ox,oy,oz).rsp ) {
                        isMaximum = false;
                        break;
                    }
                }

                // Method II: Based on Sigma
                //int sigma = (int) ceil( src.at(x,y,z).sigma );
                //for( int i = -sigma; i <= sigma; i++ ) for( int j = -sigma; j <= sigma; j++ ) {
                //	Vec3i offset = src.at(x,y,z).normals[0] * i + src.at(x,y,z).normals[1] * j;
                //	int ox = x + offset[0];
                //	int oy = y + offset[1];
                //	int oz = z + offset[2];
                //	if( src.isValid(ox,oy,oz) && src.at(x,y,z).rsp < src.at(ox,oy,oz).rsp ) {
                //		isMaximum = false; break;
                //	}
                //}

                if( isMaximum ) {
                    dst.at(x,y,z) = src.at(x,y,z);
                }
            }
        }
    }
}


void ImageProcessing::edge_tracing( Data3D<Vesselness_Sig>& src, Data3D<Vesselness_Sig>& dst, const float& thres1, const float& thres2 )
{
    Data3D<float> src1d;
    src.copyDimTo( src1d, 0 );
    IP::normalize( src1d, 1.0f );

    int x, y, z;
    std::queue<Vec3i> q;
    Data3D<unsigned char> mask( src.get_size() );
    for(z=0; z<src.SZ(); z++) for (y=0; y<src.SY(); y++) for(x=0; x<src.SX(); x++) {
                if( src1d.at(x,y,z) > thres1 ) {
                    q.push( Vec3i(x,y,z) );
                    mask.at(x,y,z) = 255;
                }
            }

    Vec3i dir[26];
    for( int i=0; i<26; i++ ) {
        int index = (i + 14) % 27;
        dir[i][0] = index/9%3 - 1;
        dir[i][1] = index/3%3 - 1;
        dir[i][2] = index/1%3 - 1;
    }

    while( !q.empty() ) {
        Vec3i pos = q.front();
        q.pop();
        Vec3i off_pos;
        for( int i=0; i<26; i++ ) {
            off_pos = pos + dir[i];
            if( src.isValid(off_pos) && !mask.at( off_pos ) && src1d.at(off_pos) > thres2 ) {
                mask.at( off_pos ) = 255;
                q.push( off_pos );
            }
        }
    }

    dst.reset( src.get_size() );
    for(z=0; z<src.SZ(); z++) for (y=0; y<src.SY(); y++) for(x=0; x<src.SX(); x++) {
                if( mask.at( x,y,z ) ) {
                    dst.at( x,y,z ) = src.at( x,y,z );
                    // dst.at( x,y,z ).rsp = sqrt( src1d.at( x,y,z ) );
                }
            }
}
