#pragma once

#include <vector>
#include "data3D.h"
#include "kernel3D.h"
#include "smart_assert.h"


class Vesselness;
class Vesselness_Sig;
class Vesselness_Nor;
class Vesselness_All;

namespace ImageProcessing {
template<typename T1, typename T2>
bool GaussianBlur3D( const Data3D<T1>& src, Data3D<T2>& dst, int ksize, double sigma = 0.0 );

template<typename T>
void normalize( Data3D<T>& data, T norm_max = 1);


///////////////////////////////////////////////////////////////////////////
// NON-MAXIMUM SUPPRESSION
void non_max_suppress( const Data3D<Vesselness_Sig>& src, Data3D<Vesselness_Sig>& dst );
void edge_tracing( Data3D<Vesselness_Sig>& src, Data3D<Vesselness_Sig>& dst, const float& thres1 = 0.85f, const float& thres2 = 0.10f );
void dir_tracing( Data3D<Vesselness_All>& src, Data3D<Vesselness_Sig>& res_dt );
void edge_tracing_mst( Data3D<Vesselness_All>& src, Data3D<Vesselness_Sig>& dst, const float& thres1 = 0.85f, const float& thres2 = 0.10f );
}

namespace IP = ImageProcessing;



template<typename T1, typename T2>
bool ImageProcessing::GaussianBlur3D( const Data3D<T1>& src, Data3D<T2>& dst, int ksize, double sigma )
{
    smart_return( ksize%2!=0, "kernel size should be odd number", false );

    //////////////////////////////////////////////////////////////////////////////////////
    // Relationship between Sigma and Kernal Size (ksize)
    /////////////////////////////
    // Option I - Based on OpenCV
    //   Calculate sigma from size:
    //         sigma = 0.3 * ( (ksize-1)/2 - 1 ) + 0.8 = 0.15*ksize + 0.35
    //   Calculate size from sigma:
    //         ksize = ( sigma - 0.35 ) / 0.15 = 6.67 * sigma - 2.33
    //   Reference: OpenCv Documentation 2.6.4.0 getGaussianKernel
    //     http://docs.opencv.org/modules/imgproc/doc/filtering.html#creategaussianfilter
    // Option II - Based on the traditional 99.7%
    //   Calculate size from sigma:
    //         ksize = 6 * sigma + 1
    //   Calculate sigma from ksize:
    //         sigma = (size - 1)/6 = 0.17 * size - 0.17

    cv::Mat gk = cv::getGaussianKernel( ksize, sigma, CV_64F );

    int hsize = ksize/2;
    smart_assert( 2*hsize+1==ksize, "Alert: Bug!" );

    cv::Vec3i spos, epos;
    spos = cv::Vec3i(0, 0, 0);
    epos = src.get_size();

    // gaussian on x-direction
    Data3D<T2> tmp1( src.get_size(), T2(0) ); // the data will be set to zero
    #pragma omp parallel for
    for( int z=spos[2]; z<epos[2]; z++ ) for( int y=spos[1]; y<epos[1]; y++ ) for( int x=spos[0]; x<epos[0]; x++ ) {
                double sum = 0.0;
                for( int i=0; i<ksize; i++) {
                    int x2 = x+i-hsize;
                    if( x2>=0 && x2<epos[0] ) {
                        tmp1.at(x, y, z) = T2( gk.at<double>(i) * src.at(x2, y, z) + tmp1.at(x, y, z));
                        sum += gk.at<double>(i);
                    }
                }
                tmp1.at(x, y, z) = T2( tmp1.at(x, y, z)/sum );
            }


    // gaussian on y-direction
    Data3D<T2> tmp2( src.get_size(), T2(0) ); // the data will be set to zero
    #pragma omp parallel for
    for( int z=spos[2]; z<epos[2]; z++ ) for( int y=spos[1]; y<epos[1]; y++ ) for( int x=spos[0]; x<epos[0]; x++ ) {
                double sum = 0.0;
                for( int i=0; i<ksize; i++) {
                    int y2 = y+i-hsize;
                    if( y2>=0 && y2<epos[1] ) {
                        tmp2.at(x, y, z) = T2( gk.at<double>(i) * tmp1.at(x, y2, z) + tmp2.at(x, y, z));
                        sum += gk.at<double>(i);
                    }
                }
                tmp2.at(x, y, z) = T2( tmp2.at(x, y, z)/sum );
            }

    tmp1.reset(); // tmp1 is no long in use. release memory

    // gaussian on z-direction
    dst.reset( src.get_size() );
    #pragma omp parallel for
    for( int z=spos[2]; z<epos[2]; z++ ) for( int y=spos[1]; y<epos[1]; y++ ) for( int x=spos[0]; x<epos[0]; x++ ) {
                double sum = 0.0;
                for( int i=0; i<ksize; i++) {
                    int z2 = z+i-hsize;
                    if( z2>=0 && z2<epos[2] ) {
                        dst.at(x, y, z) = T2( dst.at(x, y, z) + gk.at<double>(i) * tmp2.at(x, y, z2) );
                        sum += gk.at<double>(i);
                    }
                }
                dst.at(x, y, z) = T2( dst.at(x, y, z)/sum );
            }

    return true;
}


template<typename T>
void ImageProcessing::normalize( Data3D<T>& data, T norm_max )
{
    cv::Vec<T, 2> min_max = data.get_min_max_value();
    data.getMat() = 1.0 * norm_max / (min_max[1]-min_max[0]) * (data.getMat() - min_max[0]);
}
